\CAST.XPL	12-Jan-2006	Loren Blaney	loren_blaney@idcomm.com
\Ray Caster - Use the mouse to move around
\Compile with 16-bit XPL (xx.bat)

inc	c:\cxpl\codesi;		\intrinsic code declarations

def	Barrier= 20;		\how close the viewer can get to a wall

int	CpuReg,			\address of CPU register array (from GetReg)
	Cell,			\2-dimensional array containing the floor plan
	MouseDx, MouseDy,	\amount mouse moved since last GetMouseMove call
	Vx, Vy;			\viewer's coordinates in the floor plan
real	Vd;			\viewer's direction (radians: 0=east pi/2=south)

char	SpWall(14, 2+64*64);	\wall panel images (sprites)
seg char Image(1);		\buffer for copy of video screen

def	Pi = 3.141592654,
	PiOver2 = Pi/2.0,	\90 degrees
	Deg2Rad = Pi/180.0;	\multiplier to convert degrees to radians



proc	Exit(Msg);		\Display (error) message and terminate program
char	Msg;
begin
SetVid(3);			\restore normal text mode
Text(0, Msg);
Crlf(0);
OpenI(0);			\discard any keystrokes
if Geterr then [];		\discard any pending runtime errors
exit;
end;	\Exit



func	Trunc(X);		\Convert real to integer by truncation
real	X;
return Fix(X-0.5);

\=============================== MOUSE ROUTINES ================================

func	OpenMouse;		\Initialize mouse; return 'false' if it fails
begin				\pointer is set to center of screen but is hidden
CpuReg(0):= $3533;		\make sure mouse vector ($33) points to something
SoftInt($21);
if CpuReg(1)=0 & CpuReg(11)=0 then return false;

CpuReg(0):= $0000;		\reset mouse
SoftInt($33);
return CpuReg(0)		\return 'false' if failure
end;	\OpenMouse



proc	GetMouseMove;		\Get amount mouse moved since last call
begin
Cpureg(0):= $000B;
SoftInt($33);
MouseDx:= Cpureg(2);
MouseDy:= Cpureg(3);
end;	\GetMouseMove

\================================ IMAGE LOADING ================================

proc	LoadWalls;	\Load sprite images from .BMP file and set up color regs
int	Hand,		\file handle
	W, H,		\dimensions of entire image in .BMP file (pixels)
	R, G, B,	\red, green, blue
	X, Y,		\screen coordinates (pixels)
	Y320,		\Y*320 (for efficiency)
	I, T;		\index and temporary scratch
seg char Screen(1);	\temporary memory area for loading images



proc	LoadWall(X0, Y0, W, H, Sp);	\Load a sprite data array from Screen
int	X0, Y0,		\coordinates of sprite in Screen (pixels)
	W, H;		\width and height of sprite (pixels)
char	Sp;		\sprite to load
int	I, X, Y;
begin
Sp(0):= W;
Sp(1):= H;
I:= 2;
for Y:= Y0, Y0+H-1 do
    for X:= X0, X0+W-1 do
	begin
	Sp(I):= Screen(0, X+Y*320);
	I:= I + 1;
	end;
end;	\LoadWall



begin	\LoadWalls
Screen(0):= Malloc(320/16*200);	\temporary 64000-byte hunk of memory

\Read in a 256-color .BMP file
Trap(false);			\don't trap "file not found" error; we'll do it
Hand:= FOpen("WALLS.BMP", 0);	\open file for input
if Geterr \#0\ then Exit("WALLS.BMP not found");
Trap(true);
FSet(Hand, ^I);			\set device 3 to file handle
OpenI(3);

for I:= 0, 17 do T:= Chin(3);	\skip unused header info
W:= Chin(3) + Chin(3)<<8;	\0..32767 (is ample range)
for I:= 0, 2-1 do T:= Chin(3);	\skip
H:= Chin(3) + Chin(3)<<8;	\0..32767 (is ample range)
for I:= 24, 53 do T:= Chin(3);	\skip

port($3C8):= 0;			\set color registers
for I:= 0, 255 do
	begin
	B:= Chin(3)>>2;
	G:= Chin(3)>>2;
	R:= Chin(3)>>2;
	T:= Chin(3);
	port($3C9):= R;
	port($3C9):= G;
	port($3C9):= B;
	end;

\Load .BMP image into Screen array
for Y:= -(H-1), 0 do		\.BMP files are upside down (!)
	begin
	Y320:= 320 * -Y;
	for X:= 0, W-1 do
		Screen(0, X+Y320):= Chin(3);
	end;
FClose(Hand);			\close handle (so it could be used again)

for I:= 0, 4 do
	begin
	LoadWall(I*64,  0, 64, 64, addr SpWall(I,0));
	LoadWall(I*64, 64, 64, 64, addr SpWall(I+5,0));
	end;
for I:= 0, 3 do
	LoadWall(I*64,128, 64, 64, addr SpWall(I+10,0));

Release(Screen(0));		\(for neatness, so memory can be reused)
end;	\LoadWalls

\================================== THE MEAT ===================================

proc	DrawWalls;	\Draw a series of wall panels forming the rooms
int	X, Y,		\coordinates in Image array (pixels)
	Y320,		\Y*320 (for efficiency)
	CBx, CBy,	\cell boundary coordinates
	Wy,		\coordinates in wall (Y direction is vertical)
	Wx1,		\X offset into a wall panel
	WP,		\wall panel code (0..$F)
	Dist1,		\modified distance, used to eliminate fish-eye effect
	Color;		\color of pixel that gets displayed
real	Wx,		\coordinates in wall (Y direction is vertical)
	Vx1, Vy1,	\coordinates of start of ray (hops from cell to cell)
	VA,		\view angle of ray (in radians) relative to viewer (Vd)
	Dv, Dh,		\distances to vertical & horizontal cell boundaries
	Dist,		\distance from viewer to point where ray hits wall
	SinA, CosA;	\sine and cosine of view angle (for efficiency)
int	Face, FaceV, FaceH;
def	N=12, W=8, S=4, E=0;

begin
for X:= 0, 320-1 do	\raster fill Image (up & down instead of left & right)
	begin
	\Sweep projected ray from -30 to +30 degrees (sweep from left to right)
	VA:= Atan2(Float(X-160), 277.2);

	\Given viewer's position and ray's direction find Wx and Dist:
	CosA:= Cos(VA+Vd);
	SinA:= Sin(VA+Vd);
	Vx1:= Float(Vx);		\working copy of ray's starting position
	Vy1:= Float(Vy);
	repeat	begin			\loop thru cells until ray hits a wall
		if CosA >= 0.0 then			\ray heading toward East
			begin
			CBx:= (Trunc(Vx1)+64) & ~$3F;	\cell's eastern boundary
			FaceV:= W;
			end
		else	begin
			CBx:= (Trunc(Vx1) & ~$3F) - 1;	\cell's western boundary
			FaceV:= E;
			end;
		Dv:= (Float(CBx)-Vx1)/CosA;		\distance to boundary

		if SinA >= 0.0 then			\ray heading Southward
			begin
			CBy:= (Trunc(Vy1)+64) & ~$3F;	\cell's south boundary
			FaceH:= N;
			end
		else	begin
			CBy:= (Trunc(Vy1) & ~$3F) - 1;	\cell's north boundary
			FaceH:= S;
			end;
		Dh:= (Float(CBy)-Vy1)/SinA;		\distance to boundary

		if Dv <= Dh then		\use the shorter distance
			begin			\toward vertical cell boundary
			Wx:= Vy1 + Dv*SinA;	\intercept on vertical boundary
			Vx1:= Float(CBx);	\continue ray thru next cell
			Vy1:= Wx;
			Face:= FaceV;
			end
		else	begin			\toward horizontal cell boundary
			Wx:= Vx1 + Dh*CosA;	\intercept on horz. boundary
			Vx1:= Wx;		\continue ray thru next cell
			Vy1:= Float(CBy);
			Face:= FaceH;
			end;

		WP:= Cell(Trunc(Vx1)>>6, Trunc(Vy1)>>6);\wall panel of next cell
		end;
	until	WP \#0\;			\ray struck a wall

	Dv:= Vx1 - Float(Vx);			\distance from viewer to wall
	Dh:= Vy1 - Float(Vy);
	Dist:= Sqrt(Dv*Dv + Dh*Dh);
	Dist1:= Fix(Dist * Cos(VA)) /2;		\eliminate fish-eye effect
						\ (use parallel projection)
	Wx1:= Fix(Wx) & $3F;			\horz. position in wall panel
	WP:= WP>>Face & $000F;

	Y320:= 0;			\set up column of pixels to be displayed
	for Y:= 0, 200-1 do
		begin			\speed-critical code: 64000 per frame
		Wy:= ((Y-100)*Dist1)/128 + 32;	\256 = height scale factor
		case of				\100=200/2; 32=64/2; (centering)
		  Wy<0:  Color:= 7;				\ceiling
		  Wy>63: Color:= $A4				\floor
		other	 Color:= SpWall(WP-1, 2+Wx1+Wy<<6);	\wall panel
		Image(0, X+Y320):= Color;
		Y320:= Y320 + 320;
		end;
	end;
end;	\DrawWalls

\-------------------------------------------------------------------------------

proc	MoveViewer;
\Move viewer's position (Vx,Vy) and direction (Vd) using the mouse
\ -Y axis moves forward and back
\  X axis moves clockwise and counter-clockwise
real	A, D, DMax;
int	Vx6, Vy6,
	CBx, CBy;	\cell boundaries
begin
GetMouseMove;				\get change in mouse position
A:= Float(MouseDx) / 1000.0;		\X-axis motion is change in direction
Vd:= Vd + A;
D:= -Float(MouseDy/3);			\Y-axis motion moves along curent dir
DMax:= Float(Barrier-1);		\speed limit prevents moving thru walls
if D > DMax then D:= DMax;
if D < -DMax then D:= -DMax;

\Move viewer and limit how close she can get to a cell containing a block
Vx:= Vx + Fix(D*Cos(Vd));		\make tenative East-West move
Vx6:= Vx>>6;
Vy6:= Vy>>6;

CBx:= (Vx+64) & ~$3F;
if Cell(Vx6+1, Vy6) \#0\ then		\cell to the East contains a block
	if Vx > CBx-Barrier then Vx:= CBx-Barrier;
CBy:= (Vy+64) & ~$3F;
if Cell(Vx6+1, Vy6+1) \#0\ then		\cell to the SE contains a block
	if Vx>CBx-Barrier & Vy>CBy-Barrier then Vx:= CBx-Barrier;
CBy:= Vy & ~$3F;
if Cell(Vx6+1, Vy6-1) \#0\ then		\cell to the NE contains a block
	if Vx>CBx-Barrier & Vy<CBy+Barrier then Vx:= CBx-Barrier;

CBx:= Vx & ~$3F;
if Cell(Vx6-1, Vy6) \#0\ then		\cell to the West contains a block
	if Vx < CBx+Barrier then Vx:= CBx+Barrier;
CBy:= (Vy+64) & ~$3F;
if Cell(Vx6-1, Vy6+1) \#0\ then		\cell to the SW contains a block
	if Vx<CBx+Barrier & Vy>CBy-Barrier then Vx:= CBx+Barrier;
CBy:= Vy & ~$3F;
if Cell(Vx6-1, Vy6-1) \#0\ then		\cell to the NW contains a block
	if Vx<CBx+Barrier & Vy<CBy+Barrier then Vx:= CBx+Barrier;

Vy:= Vy + Fix(D*Sin(Vd));
Vx6:= Vx>>6;
Vy6:= Vy>>6;

CBy:= (Vy+64) & ~$3F;
if Cell(Vx6, Vy6+1) \#0\ then		\cell to the South contains a block
	if Vy > CBy-Barrier then Vy:= CBy-Barrier;
CBx:= (Vx+64) & ~$3F;
if Cell(Vx6+1, Vy6+1) \#0\ then		\cell to the SE contains a block
	if Vy>CBy-Barrier & Vx>CBx-Barrier then Vy:= CBy-Barrier;
CBx:= Vx & ~$3F;
if Cell(Vx6-1, Vy6+1) \#0\ then		\cell to the SW contains a block
	if Vy>CBy-Barrier & Vx<CBx+Barrier then Vy:= CBy-Barrier;

CBy:= Vy & ~$3F;
if Cell(Vx6, Vy6-1) \#0\ then		\cell to the North contains a block
	if Vy < CBy+Barrier then Vy:= CBy+Barrier;
CBx:= (Vx+64) & ~$3F;
if Cell(Vx6+1, Vy6-1) \#0\ then		\cell to the NE contains a block
	if Vy<CBy+Barrier & Vx>CBx-Barrier then Vy:= CBy+Barrier;
CBx:= Vx & ~$3F;
if Cell(Vx6-1, Vy6-1) \#0\ then		\cell to the NW contains a block
	if Vy<CBy+Barrier & Vx<CBx+Barrier then Vy:= CBy+Barrier;
end;	\MoveViewer

\===============================================================================

begin	\Main
\Array of cells describing the floor plan:
\Each cell is either open (=0) or occupied by a block.
\Each block has 4 walls.
\Each wall is specified by a hex digit 1..F.
\The direction a wall faces is specified by the order of the digits: N,W,S,E.
\For example, the block $1234 has wall image number 2 facing West.
Cell:=				\West                            --------> Y
    [[$5555,$5555,$5555,$555A,$5555,$5555,$5555,$1111,$1112,$1111,$5555], \   |
     [$55B5,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$B555], \   |
     [$55C5,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$C555], \   |
     [$55D5,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$D555], \   |
     [$5555,$1111,    0,$1113,$1111,    0,$1111,    0,    0,    0,$C555], \   v
     [$5585,    0,    0,    0,$3111,    0,$1111,    0,    0,    0,$C555], \   X
     [$5595,    0,    0,    0,$3111,    0,    0,    0,    0,    0,$C555],
     [$5565,    0,    0,    0,$3111,    0,$1111,    0,    0,    0,$9555],
\N\  [$5575,    0,    0,    0,$4111,    0,$1111,$3434,$3333,$3333,$5555], \South
     [$5565,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$6555],
     [$5575,    0,    0,    0,$1111,    0,    0,    0,    0,    0,$7555],
     [$5555,$3134,$3133,$3133,$1111,$1111,$1111,    0,    0,    0,$6555],
     [$5766,    0,    0,    0,$3111,    0,    0,    0,    0,    0,$E555],
     [    0,    0,$1111,    0,    0,    0,    0,    0,    0,    0,$6555],
     [$5675,    0,    0,    0,$1111,    0,    0,    0,    0,$5A55,$5555],
     [$5555,$5555,$5555,$5A55,$5555,$5555,$5555,$5A55,$5555,$5555,    0]];
				\East
CpuReg:= GetReg;
if not OpenMouse then Exit("This program requires a mouse.");
SetVid($13);				\320x200 with 256 colors
LoadWalls;

Image(0):= Malloc(320/16*200);		\64000 bytes
Trap(false);				\ignore div-by-0 error
Vx:= 64+Barrier;   Vy:= 64+Barrier;	\start viewer in north-west corner
Vd:= PiOver2/2.0;			\face south-east

GetMouseMove;				\initialize mouse's starting position
loop	begin
	MoveViewer;
	DrawWalls;			\draw into Image array
	repeat until port($3DA) & $08;	\wait for vertical blank

	\Copy Image to screen (takes about 5ms--outruns video beam--no glitches)
	Blit(Image(0), 0, $A000, 0, 64*1000);
	if ChkKey then Exit(" ");	\terminate program when a key is hit
	end;
end;	\Main
